# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Literal, Self, cast

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import exceptions
from openstack import resource
from openstack import utils


class MetadefNamespace(resource.Resource, tag.TagMixin):
    resources_key = 'namespaces'
    base_path = '/metadefs/namespaces'

    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_list = True
    allow_delete = True

    _query_mapping = resource.QueryParameters(
        "limit",
        "marker",
        "resource_types",
        "sort_dir",
        "sort_key",
        "visibility",
    )

    created_at = resource.Body('created_at')
    description = resource.Body('description')
    display_name = resource.Body('display_name')
    is_protected = resource.Body('protected', type=bool)
    namespace = resource.Body('namespace', alternate_id=True)
    owner = resource.Body('owner')
    resource_type_associations = resource.Body(
        'resource_type_associations',
        type=list,
        list_type=dict,
    )
    updated_at = resource.Body('updated_at')
    visibility = resource.Body('visibility')

    def _commit(
        self,
        session: adapter.Adapter,
        request: resource._Request,
        method: Literal['POST', 'PUT', 'PATCH'],
        microversion: str | None = None,
        has_body: bool = True,
        retry_on_conflict: bool | None = None,
    ) -> Self:
        # Rather annoyingly, Glance insists on us providing the 'namespace'
        # argument, even if we're not changing it. We need to add this here
        # since it won't be included if Resource.commit thinks its unchanged
        # TODO(stephenfin): Eventually we could indicate attributes that are
        # required in the body on update, like 'CreateOpts.requires_id' does
        # for the ID in the URL
        cast(dict[str, Any], request.body)['namespace'] = self.namespace

        return super()._commit(
            session,
            request,
            method,
            microversion,
            has_body=True,
            retry_on_conflict=None,
        )

    def _delete_all(self, session: adapter.Adapter, url: str) -> Self:
        response = session.delete(url)
        exceptions.raise_from_response(response)
        self._translate_response(response, has_body=False)
        return self

    def delete_all_properties(self, session: adapter.Adapter) -> Self:
        """Delete all properties in a namespace.

        :param session: The session to use for making this request
        :returns: The server response
        """

        url = utils.urljoin(self.base_path, self.id, 'properties')
        return self._delete_all(session, url)

    def delete_all_objects(self, session: adapter.Adapter) -> Self:
        """Delete all objects in a namespace.

        :param session: The session to use for making this request
        :returns: The server response
        """
        url = utils.urljoin(self.base_path, self.id, 'objects')
        return self._delete_all(session, url)

    # NOTE(mrjoshi): This method is re-implemented as we require a ``POST``
    # call while the original method does a ``PUT`` call.
    def add_tag(self, session: resource.AdapterT, tag: str) -> Self:
        """Adds a single tag to the resource.

        :param session: The session to use for making this request.
        :param tag: The tag as a string.
        """
        url = utils.urljoin(self.base_path, self.id, 'tags', tag)
        session = self._get_session(session)
        response = session.post(url)
        exceptions.raise_from_response(response)
        # we do not want to update tags directly
        tags = self.tags
        tags.append(tag)
        self._body.attributes.update({'tags': tags})
        return self

    # NOTE(mrjoshi): This method is re-implemented to add support for the
    # 'append' option. This method uses a ``POST`` call rather than the
    # standard ``PUT`` call.
    def set_tags(
        self, session: resource.AdapterT, tags: list[str], append: bool = False
    ) -> Self:
        """Sets/Replaces all tags on the resource.

        :param session: The session to use for making this request.
        :param list tags: List with tags to be set on the resource
        :param append: If set to true, adds new tags to existing tags,
            else overwrites the existing tags with new ones.
        """
        url = utils.urljoin(self.base_path, self.id, 'tags')
        session = self._get_session(session)

        headers = {'X-OpenStack-Append': 'False'}
        if append:
            headers['X-Openstack-Append'] = 'True'

        response = session.post(
            url, headers=headers, json={'tags': [{'name': x} for x in tags]}
        )
        exceptions.raise_from_response(response)

        self._body.attributes.update({'tags': tags})

        return self



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.instance_ha.v1 import _proxy
from openstack.instance_ha.v1 import host
from openstack.instance_ha.v1 import notification
from openstack.instance_ha.v1 import segment
from openstack.instance_ha.v1 import vmove
from openstack.tests.unit import test_proxy_base

SEGMENT_ID = "c50b96eb-2a66-40f8-bca8-c5fa90d595c0"
HOST_ID = "52d05e43-d08e-42b8-ae33-e47c8ea2ad47"
NOTIFICATION_ID = "a0e70d3a-b3a2-4616-b65d-a7c03a2c85fc"
VMOVE_ID = "16a7c91f-8342-49a7-c731-3a632293f845"


class TestInstanceHaProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestInstanceHaHosts(TestInstanceHaProxy):
    def test_hosts(self):
        self.verify_list(
            self.proxy.hosts,
            host.Host,
            method_args=[SEGMENT_ID],
            expected_args=[],
            expected_kwargs={"segment_id": SEGMENT_ID},
        )

    def test_host_get(self):
        self.verify_get(
            self.proxy.get_host,
            host.Host,
            method_args=[HOST_ID],
            method_kwargs={"segment_id": SEGMENT_ID},
            expected_kwargs={"segment_id": SEGMENT_ID},
        )

    def test_host_create(self):
        self.verify_create(
            self.proxy.create_host,
            host.Host,
            method_args=[SEGMENT_ID],
            method_kwargs={},
            expected_args=[],
            expected_kwargs={"segment_id": SEGMENT_ID},
        )

    def test_host_update(self):
        self.verify_update(
            self.proxy.update_host,
            host.Host,
            method_kwargs={"segment_id": SEGMENT_ID},
        )

    def test_host_delete(self):
        self.verify_delete(
            self.proxy.delete_host,
            host.Host,
            True,
            method_kwargs={"segment_id": SEGMENT_ID},
            expected_kwargs={"segment_id": SEGMENT_ID},
        )


class TestInstanceHaNotifications(TestInstanceHaProxy):
    def test_notifications(self):
        self.verify_list(self.proxy.notifications, notification.Notification)

    def test_notification_get(self):
        self.verify_get(self.proxy.get_notification, notification.Notification)

    def test_notification_create(self):
        self.verify_create(
            self.proxy.create_notification, notification.Notification
        )


class TestInstanceHaSegments(TestInstanceHaProxy):
    def test_segments(self):
        self.verify_list(self.proxy.segments, segment.Segment)

    def test_segment_get(self):
        self.verify_get(self.proxy.get_segment, segment.Segment)

    def test_segment_create(self):
        self.verify_create(self.proxy.create_segment, segment.Segment)

    def test_segment_update(self):
        self.verify_update(self.proxy.update_segment, segment.Segment)

    def test_segment_delete(self):
        self.verify_delete(self.proxy.delete_segment, segment.Segment, True)


class TestInstanceHaVMoves(TestInstanceHaProxy):
    def test_vmoves(self):
        self.verify_list(
            self.proxy.vmoves,
            vmove.VMove,
            method_args=[NOTIFICATION_ID],
            expected_args=[],
            expected_kwargs={"notification_id": NOTIFICATION_ID},
        )

    def test_vmove_get(self):
        self.verify_get(
            self.proxy.get_vmove,
            vmove.VMove,
            method_args=[VMOVE_ID, NOTIFICATION_ID],
            expected_args=[VMOVE_ID],
            expected_kwargs={"notification_id": NOTIFICATION_ID},
        )



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from unittest import mock

from openstack.baremetal import configdrive
from openstack.tests.unit import base


class TestPopulateDirectory(base.TestCase):
    def _check(
        self, metadata, user_data=None, network_data=None, vendor_data=None
    ):
        with configdrive.populate_directory(
            metadata,
            user_data=user_data,
            network_data=network_data,
            vendor_data=vendor_data,
        ) as d:
            for version in ('2012-08-10', 'latest'):
                with open(
                    os.path.join(d, 'openstack', version, 'meta_data.json')
                ) as fp:
                    actual_metadata = json.load(fp)

                self.assertEqual(metadata, actual_metadata)
                network_data_file = os.path.join(
                    d, 'openstack', version, 'network_data.json'
                )
                user_data_file = os.path.join(
                    d, 'openstack', version, 'user_data'
                )
                vendor_data_file = os.path.join(
                    d, 'openstack', version, 'vendor_data2.json'
                )

                if network_data is None:
                    self.assertFalse(os.path.exists(network_data_file))
                else:
                    with open(network_data_file) as fp:
                        self.assertEqual(network_data, json.load(fp))

                if vendor_data is None:
                    self.assertFalse(os.path.exists(vendor_data_file))
                else:
                    with open(vendor_data_file) as fp:
                        self.assertEqual(vendor_data, json.load(fp))

                if user_data is None:
                    self.assertFalse(os.path.exists(user_data_file))
                else:
                    if isinstance(user_data, str):
                        user_data = user_data.encode()
                    with open(user_data_file, 'rb') as fp:
                        self.assertEqual(user_data, fp.read())

        # Clean up in __exit__
        self.assertFalse(os.path.exists(d))

    def test_without_user_data(self):
        self._check({'foo': 42})

    def test_with_user_data(self):
        self._check({'foo': 42}, b'I am user data')

    def test_with_user_data_as_string(self):
        self._check({'foo': 42}, 'I am user data')

    def test_with_network_data(self):
        self._check({'foo': 42}, network_data={'networks': {}})

    def test_with_vendor_data(self):
        self._check({'foo': 42}, vendor_data={'foo': 'bar'})


@mock.patch('subprocess.Popen', autospec=True)
class TestPack(base.TestCase):
    def test_no_genisoimage(self, mock_popen):
        mock_popen.side_effect = OSError
        self.assertRaisesRegex(
            RuntimeError, "genisoimage", configdrive.pack, "/fake"
        )

    def test_genisoimage_fails(self, mock_popen):
        mock_popen.return_value.communicate.return_value = b"", b"BOOM"
        mock_popen.return_value.returncode = 1
        self.assertRaisesRegex(RuntimeError, "BOOM", configdrive.pack, "/fake")

    def test_success(self, mock_popen):
        mock_popen.return_value.communicate.return_value = b"", b""
        mock_popen.return_value.returncode = 0
        result = configdrive.pack("/fake")
        # Make sure the result is string on all python versions
        self.assertIsInstance(result, str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute.v2 import image as _image
from openstack.tests.functional.compute.v2 import base
from openstack.tests.functional.image.v2.test_image import TEST_IMAGE_NAME


class TestImage(base.BaseComputeTest):
    def setUp(self):
        super().setUp()

        # get a non-test image to work with
        images = self.admin_compute_client.images()
        self.image = next(images)

        if self.image.name == TEST_IMAGE_NAME:
            self.image = next(images)

    def test_image(self):
        # list all images

        images = list(self.admin_compute_client.images())
        self.assertGreater(len(images), 0)
        for image in images:
            self.assertIsInstance(image.id, str)

        # find image by name

        image = self.admin_compute_client.find_image(
            self.image.name, ignore_missing=False
        )
        self.assertIsInstance(image, _image.Image)
        self.assertEqual(self.image.id, image.id)
        self.assertEqual(self.image.name, image.name)

        # get image by ID

        image = self.admin_compute_client.get_image(self.image.id)
        self.assertIsInstance(image, _image.Image)
        self.assertEqual(self.image.id, image.id)
        self.assertEqual(self.image.name, image.name)

    def test_image_metadata(self):
        # delete pre-existing metadata

        self.admin_compute_client.delete_image_metadata(
            self.image, self.image.metadata.keys()
        )
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertFalse(image.metadata)

        # get metadata (should be empty)

        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertFalse(image.metadata)

        # set no metadata

        self.admin_compute_client.set_image_metadata(self.image)
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertFalse(image.metadata)

        # set empty metadata

        self.admin_compute_client.set_image_metadata(self.image, k0='')
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertIn('k0', image.metadata)
        self.assertEqual('', image.metadata['k0'])

        # set metadata

        self.admin_compute_client.set_image_metadata(self.image, k1='v1')
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertTrue(image.metadata)
        self.assertEqual(2, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1', image.metadata['k1'])

        # set more metadata

        self.admin_compute_client.set_image_metadata(self.image, k2='v2')
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertTrue(image.metadata)
        self.assertEqual(3, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1', image.metadata['k1'])
        self.assertIn('k2', image.metadata)
        self.assertEqual('v2', image.metadata['k2'])

        # update metadata

        self.admin_compute_client.set_image_metadata(self.image, k1='v1.1')
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertTrue(image.metadata)
        self.assertEqual(3, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1.1', image.metadata['k1'])
        self.assertIn('k2', image.metadata)
        self.assertEqual('v2', image.metadata['k2'])

        # delete all metadata (cleanup)

        self.admin_compute_client.delete_image_metadata(
            self.image, image.metadata.keys()
        )
        image = self.admin_compute_client.fetch_image_metadata(self.image)
        self.assertFalse(image.metadata)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.network.v2 import security_group
from openstack.network.v2 import security_group_rule
from openstack.tests.functional import base


class TestSecurityGroupRule(base.BaseFunctionalTest):
    IPV4 = "IPv4"
    PROTO = "tcp"
    PORT = 22
    DIR = "ingress"
    ID: str
    RULE_ID: str

    def setUp(self):
        super().setUp()
        self.NAME = self.getUniqueString()
        sot = self.user_cloud.network.create_security_group(name=self.NAME)
        assert isinstance(sot, security_group.SecurityGroup)
        self.assertEqual(self.NAME, sot.name)
        self.ID = sot.id
        rul = self.user_cloud.network.create_security_group_rule(
            direction=self.DIR,
            ethertype=self.IPV4,
            port_range_max=self.PORT,
            port_range_min=self.PORT,
            protocol=self.PROTO,
            security_group_id=self.ID,
        )
        assert isinstance(rul, security_group_rule.SecurityGroupRule)
        self.assertEqual(self.ID, rul.security_group_id)
        self.RULE_ID = rul.id

    def tearDown(self):
        sot = self.user_cloud.network.delete_security_group_rule(
            self.RULE_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_security_group(
            self.ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.user_cloud.network.find_security_group_rule(
            self.RULE_ID, ignore_missing=False
        )
        self.assertEqual(self.RULE_ID, sot.id)

    def test_get(self):
        sot = self.user_cloud.network.get_security_group_rule(self.RULE_ID)
        self.assertEqual(self.RULE_ID, sot.id)
        self.assertEqual(self.DIR, sot.direction)
        self.assertEqual(self.PROTO, sot.protocol)
        self.assertEqual(self.PORT, sot.port_range_min)
        self.assertEqual(self.PORT, sot.port_range_max)
        self.assertEqual(self.ID, sot.security_group_id)

    def test_list(self):
        ids = [o.id for o in self.user_cloud.network.security_group_rules()]
        self.assertIn(self.RULE_ID, ids)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.compute.v2 import server_remote_console
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {'protocol': 'rdp', 'type': 'rdp', 'url': 'fake'}


class TestServerRemoteConsole(base.TestCase):
    def setUp(self):
        super().setUp()
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = '2.9'
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.sess = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_basic(self):
        sot = server_remote_console.ServerRemoteConsole()
        self.assertEqual('remote_console', sot.resource_key)
        self.assertEqual(
            '/servers/%(server_id)s/remote-consoles', sot.base_path
        )
        self.assertTrue(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = server_remote_console.ServerRemoteConsole(**EXAMPLE)
        self.assertEqual(EXAMPLE['url'], sot.url)

    def test_create_type_mks_old(self):
        sot = server_remote_console.ServerRemoteConsole(
            server_id='fake_server', type='webmks'
        )

        class FakeEndpointData:
            min_microversion = '2'
            max_microversion = '2.5'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()

        self.assertRaises(ValueError, sot.create, self.sess)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import limit as _limit
from openstack.tests.functional.identity.v3 import base


class TestLimit(base.BaseIdentityTest):
    def setUp(self):
        super().setUp()

        self.service_name = self.getUniqueString('service')
        self.service_type = self.getUniqueString('type')
        self.service = self.system_admin_identity_client.create_service(
            name=self.service_name,
            type=self.service_type,
        )
        self.addCleanup(
            self.system_admin_identity_client.delete_service, self.service
        )

        self.resource_name = self.getUniqueString('resource')
        self.registered_limit = (
            self.system_admin_identity_client.create_registered_limit(
                resource_name=self.resource_name,
                service_id=self.service.id,
                default_limit=10,
            )
        )
        self.addCleanup(
            self.system_admin_identity_client.delete_registered_limit,
            self.registered_limit,
        )

        self.project_name = self.getUniqueString('project')
        self.project = self.system_admin_identity_client.create_project(
            name=self.project_name,
        )
        self.addCleanup(
            self.system_admin_identity_client.delete_project, self.project
        )

        self.limit_description = self.getUniqueString('limit')

    def _delete_limit(self, limit):
        ret = self.system_admin_identity_client.delete_limit(limit)
        self.assertIsNone(ret)

    def test_limit(self):
        # create the limit

        limit = self.system_admin_identity_client.create_limit(
            resource_name=self.resource_name,
            service_id=self.service.id,
            project_id=self.project.id,
            resource_limit=50,
        )
        self.addCleanup(self._delete_limit, limit)
        self.assertIsInstance(limit, _limit.Limit)
        self.assertIsNotNone(limit.id)
        self.assertIsNone(limit.description)
        self.assertEqual(self.service.id, limit.service_id)
        self.assertEqual(self.project.id, limit.project_id)
        self.assertEqual(50, limit.resource_limit)

        # update the limit

        limit = self.system_admin_identity_client.update_limit(
            limit, description=self.limit_description
        )
        self.assertIsInstance(limit, _limit.Limit)
        self.assertEqual(self.limit_description, limit.description)

        # retrieve details of the (updated) limit by ID

        limit = self.system_admin_identity_client.get_limit(limit.id)
        self.assertIsInstance(limit, _limit.Limit)
        self.assertEqual(self.limit_description, limit.description)

        # (there's no name, so no way to retrieve by name)

        # list all limits

        limits = list(self.system_admin_identity_client.limits())
        self.assertIsInstance(limits[0], _limit.Limit)
        self.assertIn(
            self.resource_name,
            {x.resource_name for x in limits},
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, ClassVar, Literal, overload
from collections.abc import Callable, Generator

from openstack.database.v1 import database as _database
from openstack.database.v1 import flavor as _flavor
from openstack.database.v1 import instance as _instance
from openstack.database.v1 import user as _user
from openstack import proxy
from openstack import resource


class Proxy(proxy.Proxy):
    api_version: ClassVar[Literal['1']] = '1'

    _resource_registry = {
        "database": _database.Database,
        "flavor": _flavor.Flavor,
        "instance": _instance.Instance,
        "user": _user.User,
    }

    def create_database(
        self,
        instance: str | _instance.Instance,
        **attrs: Any,
    ) -> _database.Database:
        """Create a new database from attributes

        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.database.v1.database.Database`,
            comprised of the properties on the Database class.

        :returns: The results of server creation
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._create(
            _database.Database, instance_id=instance.id, **attrs
        )

    def delete_database(
        self,
        database: str | _database.Database,
        instance: str | _instance.Instance | None = None,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a database

        :param database: The value can be either the ID of a database or a
            :class:`~openstack.database.v1.database.Database` instance.
        :param instance: This parameter needs to be specified when
            an ID is given as `database`.
            It can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the database does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent database.

        :returns: ``None``
        """
        instance_id = self._get_uri_attribute(
            database, instance, "instance_id"
        )
        self._delete(
            _database.Database,
            database,
            instance_id=instance_id,
            ignore_missing=ignore_missing,
        )

    @overload
    def find_database(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: Literal[False],
    ) -> _database.Database: ...

    @overload
    def find_database(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: bool = True,
    ) -> _database.Database | None: ...

    def find_database(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: bool = True,
    ) -> _database.Database | None:
        """Find a single database

        :param name_or_id: The name or ID of a database.
        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.database.v1.database.Database` or None
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._find(
            _database.Database,
            name_or_id,
            instance_id=instance.id,
            ignore_missing=ignore_missing,
        )

    def databases(
        self,
        instance: str | _instance.Instance,
        **query: Any,
    ) -> Generator[_database.Database, None, None]:
        """Return a generator of databases

        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
            instance that the interface belongs to.
        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of database objects
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._list(_database.Database, instance_id=instance.id, **query)

    def get_database(
        self,
        database: str | _database.Database,
        instance: str | _instance.Instance | None = None,
    ) -> _database.Database:
        """Get a single database

        :param instance: This parameter needs to be specified when
            an ID is given as `database`.
            It can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param database: The value can be the ID of a database or a
            :class:`~openstack.database.v1.database.Database`
            instance.

        :returns: One :class:`~openstack.database.v1.database.Database`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_database.Database, database)

    @overload
    def find_flavor(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _flavor.Flavor: ...

    @overload
    def find_flavor(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _flavor.Flavor | None: ...

    def find_flavor(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _flavor.Flavor | None:
        """Find a single flavor

        :param name_or_id: The name or ID of a flavor.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.database.v1.flavor.Flavor` or None
        """
        return self._find(
            _flavor.Flavor, name_or_id, ignore_missing=ignore_missing
        )

    def get_flavor(self, flavor: str | _flavor.Flavor) -> _flavor.Flavor:
        """Get a single flavor

        :param flavor: The value can be the ID of a flavor or a
            :class:`~openstack.database.v1.flavor.Flavor` instance.

        :returns: One :class:`~openstack.database.v1.flavor.Flavor`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_flavor.Flavor, flavor)

    def flavors(self, **query: Any) -> Generator[_flavor.Flavor, None, None]:
        """Return a generator of flavors

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of flavor objects
        """
        return self._list(_flavor.Flavor, **query)

    def create_instance(self, **attrs: Any) -> _instance.Instance:
        """Create a new instance from attributes

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.database.v1.instance.Instance`,
            comprised of the properties on the Instance class.

        :returns: The results of server creation
        """
        return self._create(_instance.Instance, **attrs)

    def delete_instance(
        self, instance: str | _instance.Instance, ignore_missing: bool = True
    ) -> None:
        """Delete an instance

        :param instance: The value can be either the ID of an instance or a
            :class:`~openstack.database.v1.instance.Instance` instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the instance does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent instance.

        :returns: ``None``
        """
        self._delete(
            _instance.Instance, instance, ignore_missing=ignore_missing
        )

    @overload
    def find_instance(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _instance.Instance: ...

    @overload
    def find_instance(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _instance.Instance | None: ...

    def find_instance(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _instance.Instance | None:
        """Find a single instance

        :param name_or_id: The name or ID of a instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.database.v1.instance.Instance` or None
        """
        return self._find(
            _instance.Instance, name_or_id, ignore_missing=ignore_missing
        )

    def get_instance(
        self, instance: str | _instance.Instance
    ) -> _instance.Instance:
        """Get a single instance

        :param instance: The value can be the ID of an instance or a
            :class:`~openstack.database.v1.instance.Instance`
            instance.

        :returns: One :class:`~openstack.database.v1.instance.Instance`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_instance.Instance, instance)

    def instances(
        self,
        **query: Any,
    ) -> Generator[_instance.Instance, None, None]:
        """Return a generator of instances

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of instance objects
        """
        return self._list(_instance.Instance, **query)

    def update_instance(
        self, instance: str | _instance.Instance, **attrs: Any
    ) -> _instance.Instance:
        """Update a instance

        :param instance: Either the id of a instance or a
            :class:`~openstack.database.v1.instance.Instance` instance.
        :param attrs: The attributes to update on the instance represented
            by ``instance``.

        :returns: The updated instance
        """
        return self._update(_instance.Instance, instance, **attrs)

    def create_user(
        self,
        instance: str | _instance.Instance,
        **attrs: Any,
    ) -> _user.User:
        """Create a new user from attributes

        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.database.v1.user.User`,
            comprised of the properties on the User class.

        :returns: The results of server creation
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._create(_user.User, instance_id=instance.id, **attrs)

    def delete_user(
        self,
        user: str | _user.User,
        instance: str | _instance.Instance | None = None,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a user

        :param user: The value can be either the ID of a user or a
            :class:`~openstack.database.v1.user.User` instance.
        :param instance: This parameter needs to be specified when
            an ID is given as `user`.
            It can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the user does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent user.

        :returns: ``None``
        """
        instance = self._get_resource(_instance.Instance, instance)
        self._delete(
            _user.User,
            user,
            ignore_missing=ignore_missing,
            instance_id=instance.id,
        )

    @overload
    def find_user(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: Literal[False],
    ) -> _user.User: ...

    @overload
    def find_user(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: bool = True,
    ) -> _user.User | None: ...

    def find_user(
        self,
        name_or_id: str,
        instance: str | _instance.Instance,
        ignore_missing: bool = True,
    ) -> _user.User | None:
        """Find a single user

        :param name_or_id: The name or ID of a user.
        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.database.v1.user.User` or None
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._find(
            _user.User,
            name_or_id,
            instance_id=instance.id,
            ignore_missing=ignore_missing,
        )

    def users(
        self,
        instance: str | _instance.Instance,
        **query: Any,
    ) -> Generator[_user.User, None, None]:
        """Return a generator of users

        :param instance: This can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`
        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of user objects
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._list(_user.User, instance_id=instance.id, **query)

    def get_user(
        self,
        user: str | _user.User,
        instance: str | _instance.Instance | None = None,
    ) -> _user.User:
        """Get a single user

        :param user: The value can be the ID of a user or a
            :class:`~openstack.database.v1.user.User` instance.
        :param instance: This parameter needs to be specified when
            an ID is given as `database`.
            It can be either the ID of an instance
            or a :class:`~openstack.database.v1.instance.Instance`

        :returns: One :class:`~openstack.database.v1.user.User`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        instance = self._get_resource(_instance.Instance, instance)
        return self._get(_user.User, user)

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :returns: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int | float | None = 2,
        wait: int | None = 120,
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import fnmatch
import inspect
import ipaddress
import re
import socket
import uuid
import warnings

from decorator import decorator
import jmespath
import psutil

from openstack import _log
from openstack import exceptions
from openstack import warnings as os_warnings


def _dictify_resource(resource):
    if isinstance(resource, list):
        return [_dictify_resource(r) for r in resource]
    else:
        if hasattr(resource, 'toDict'):
            return resource.toDict()
        else:
            return resource


def _filter_list(data, name_or_id, filters):
    """Filter a list by name/ID and arbitrary meta data.

    :param list data: The list of dictionary data to filter. It is expected
        that each dictionary contains an 'id' and 'name' key if a value for
        name_or_id is given.
    :param string name_or_id: The name or ID of the entity being filtered. Can
        be a glob pattern, such as 'nb01*'.
    :param filters: A dictionary of meta data to use for further filtering.
        Elements of this dictionary may, themselves, be dictionaries. Example::

            {'last_name': 'Smith', 'other': {'gender': 'Female'}}

        OR

        A string containing a jmespath expression for further filtering.
        Invalid filters will be ignored.
    """
    # The logger is openstack.cloud.fmmatch to allow a user/operator to
    # configure logging not to communicate about fnmatch misses
    # (they shouldn't be too spammy, but one never knows)
    log = _log.setup_logging('openstack.fnmatch')
    if name_or_id:
        # name_or_id might already be unicode
        name_or_id = str(name_or_id)
        identifier_matches = []
        bad_pattern = False
        try:
            fn_reg = re.compile(fnmatch.translate(name_or_id))
        except re.error:
            # If the fnmatch re doesn't compile, then we don't care,
            # but log it in case the user DID pass a pattern but did
            # it poorly and wants to know what went wrong with their
            # search
            fn_reg = None
        for e in data:
            e_id = str(e.get('id', None))
            e_name = str(e.get('name', None))

            if (e_id and e_id == name_or_id) or (
                e_name and e_name == name_or_id
            ):
                identifier_matches.append(e)
            else:
                # Only try fnmatch if we don't match exactly
                if not fn_reg:
                    # If we don't have a pattern, skip this, but set the flag
                    # so that we log the bad pattern
                    bad_pattern = True
                    continue
                if (e_id and fn_reg.match(e_id)) or (
                    e_name and fn_reg.match(e_name)
                ):
                    identifier_matches.append(e)
        if not identifier_matches and bad_pattern:
            log.debug("Bad pattern passed to fnmatch", exc_info=True)
        data = identifier_matches

    if not filters:
        return data

    if isinstance(filters, str):
        warnings.warn(
            'Support for jmespath-style filters is deprecated and will be '
            'removed in a future release. Consider using dictionary-style '
            'filters instead.',
            os_warnings.RemovedInSDK60Warning,
        )
        return jmespath.search(filters, data)

    def _dict_filter(f, d):
        if not d:
            return False
        for key in f.keys():
            if key not in d:
                log.warning(
                    "Invalid filter: %s is not an attribute of %s.%s",
                    key,
                    e.__class__.__module__,
                    e.__class__.__qualname__,
                )
                # we intentionally skip this since the user was trying to
                # filter on _something_, but we don't know what that
                # _something_ was
                raise AttributeError(key)
            if isinstance(f[key], dict):
                if not _dict_filter(f[key], d.get(key, None)):
                    return False
            elif d.get(key, None) != f[key]:
                return False
        return True

    filtered = []
    for e in data:
        if _dict_filter(filters, e):
            filtered.append(e)
    return filtered


def _get_entity(cloud, resource, name_or_id, filters, **kwargs):
    """Return a single entity from the list returned by a given method.

    :param object cloud: The controller class (Example: the main OpenStackCloud
        object).
    :param string or callable resource: The string that identifies the resource
        to use to lookup the get_<>_by_id or search_<resource>s methods
        (Example: network) or a callable to invoke.
    :param string name_or_id: The name or ID of the entity being filtered or an
        object or dict. If this is an object/dict with an 'id' attr/key, we
        return it and bypass resource lookup.
    :param filters: A dictionary of meta data to use for further filtering.
        OR A string containing a jmespath expression for further filtering.
        Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"
    """

    # Sometimes in the control flow of openstacksdk, we already have an object
    # fetched. Rather than then needing to pull the name or id out of that
    # object, pass it in here and rely on caching to prevent us from making
    # an additional call, it's simple enough to test to see if we got an
    # object and just short-circuit return it.

    if hasattr(name_or_id, 'id') or (
        isinstance(name_or_id, dict) and 'id' in name_or_id
    ):
        return name_or_id

    # If a uuid is passed short-circuit it calling the
    # get_<resource_name>_by_id method
    if getattr(cloud, 'use_direct_get', False) and _is_uuid_like(name_or_id):
        get_resource = getattr(cloud, f'get_{resource}_by_id', None)
        if get_resource:
            return get_resource(name_or_id)

    search = (
        resource
        if callable(resource)
        else getattr(cloud, f'search_{resource}s', None)
    )
    if search:
        entities = search(name_or_id, filters, **kwargs)
        if entities:
            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}"
                )
            return entities[0]
    return None


def localhost_supports_ipv6():
    """Determine whether the local host supports IPv6

    We look for the all ip addresses configured to this node, and assume that
    if any of these is IPv6 address (but not loopback or link local), this host
    has IPv6 connectivity.
    """

    for ifname, if_addrs in psutil.net_if_addrs().items():
        for if_addr in if_addrs:
            if if_addr.family != socket.AF_INET6:
                continue
            addr = ipaddress.ip_address(if_addr.address)
            if not addr.is_link_local and not addr.is_loopback:
                return True
    return False


def valid_kwargs(*valid_args):
    # This decorator checks if argument passed as **kwargs to a function are
    # present in valid_args.
    #
    # Typically, valid_kwargs is used when we want to distinguish between
    # None and omitted arguments and we still want to validate the argument
    # list.
    #
    # Example usage:
    #
    # @valid_kwargs('opt_arg1', 'opt_arg2')
    # def my_func(self, mandatory_arg1, mandatory_arg2, **kwargs):
    #   ...
    #
    @decorator
    def func_wrapper(func, *args, **kwargs):
        argspec = inspect.getfullargspec(func)
        for k in kwargs:
            if k not in argspec.args[1:] and k not in valid_args:
                raise TypeError(
                    f"{inspect.stack()[1][3]}() got an unexpected keyword "
                    f"argument '{k}'"
                )
        return func(*args, **kwargs)

    return func_wrapper


@contextlib.contextmanager
def openstacksdk_exceptions(error_message=None):
    """Context manager for dealing with openstack exceptions.

    :param string error_message: String to use for the exception message
        content on non-SDKException exception.

        Useful for avoiding wrapping SDKException exceptions
        within themselves. Code called from within the context may throw such
        exceptions without having to catch and reraise them.

        Non-SDKException exceptions thrown within the context will
        be wrapped and the exception message will be appended to the given
        error message.
    """
    try:
        yield
    except exceptions.SDKException:
        raise
    except Exception as e:
        if error_message is None:
            error_message = str(e)
        raise exceptions.SDKException(error_message)


def safe_dict_min(key, data):
    """Safely find the minimum for a given key in a list of dict objects.

    This will find the minimum integer value for specific dictionary key
    across a list of dictionaries. The values for the given key MUST be
    integers, or string representations of an integer.

    The dictionary key does not have to be present in all (or any)
    of the elements/dicts within the data set.

    :param string key: The dictionary key to search for the minimum value.
    :param list data: List of dicts to use for the data set.

    :returns: None if the field was not found in any elements, or
        the minimum value for the field otherwise.
    """
    min_value = None
    for d in data:
        if (key in d) and (d[key] is not None):
            try:
                val = int(d[key])
            except ValueError:
                raise exceptions.SDKException(
                    "Search for minimum value failed. "
                    f"Value for {key} is not an integer: {d[key]}"
                )
            if (min_value is None) or (val < min_value):
                min_value = val
    return min_value


def safe_dict_max(key, data):
    """Safely find the maximum for a given key in a list of dict objects.

    This will find the maximum integer value for specific dictionary key
    across a list of dictionaries. The values for the given key MUST be
    integers, or string representations of an integer.

    The dictionary key does not have to be present in all (or any)
    of the elements/dicts within the data set.

    :param string key: The dictionary key to search for the maximum value.
    :param list data: List of dicts to use for the data set.

    :returns: None if the field was not found in any elements, or
        the maximum value for the field otherwise.
    """
    max_value = None
    for d in data:
        if (key in d) and (d[key] is not None):
            try:
                val = int(d[key])
            except ValueError:
                raise exceptions.SDKException(
                    "Search for maximum value failed. "
                    f"Value for {key} is not an integer: {d[key]}"
                )
            if (max_value is None) or (val > max_value):
                max_value = val
    return max_value


def parse_range(value):
    """Parse a numerical range string.

    Breakdown a range expression into its operater and numerical parts.
    This expression must be a string. Valid values must be an integer string,
    optionally preceeded by one of the following operators::

        - "<"  : Less than
        - ">"  : Greater than
        - "<=" : Less than or equal to
        - ">=" : Greater than or equal to

    Some examples of valid values and function return values::

        - "1024"  : returns (None, 1024)
        - "<5"    : returns ("<", 5)
        - ">=100" : returns (">=", 100)

    :param string value: The range expression to be parsed.

    :returns: A tuple with the operator string (or None if no operator
        was given) and the integer value. None is returned if parsing failed.
    """
    if value is None:
        return None

    range_exp = re.match(r'(<|>|<=|>=){0,1}(\d+)$', value)
    if range_exp is None:
        return None

    op = range_exp.group(1)
    num = int(range_exp.group(2))
    return (op, num)


def range_filter(data, key, range_exp):
    """Filter a list by a single range expression.

    :param list data: List of dictionaries to be searched.
    :param string key: Key name to search within the data set.
    :param string range_exp: The expression describing the range of values.

    :returns: A list subset of the original data set.
    :raises: :class:`~openstack.exceptions.SDKException` on invalid range
        expressions.
    """
    filtered = []
    range_exp = str(range_exp).upper()

    if range_exp == "MIN":
        key_min = safe_dict_min(key, data)
        if key_min is None:
            return []
        for d in data:
            if int(d[key]) == key_min:
                filtered.append(d)
        return filtered
    elif range_exp == "MAX":
        key_max = safe_dict_max(key, data)
        if key_max is None:
            return []
        for d in data:
            if int(d[key]) == key_max:
                filtered.append(d)
        return filtered

    # Not looking for a min or max, so a range or exact value must
    # have been supplied.
    val_range = parse_range(range_exp)

    # If parsing the range fails, it must be a bad value.
    if val_range is None:
        raise exceptions.SDKException(f"Invalid range value: {range_exp}")

    op = val_range[0]
    if op:
        # Range matching
        for d in data:
            d_val = int(d[key])
            if op == '<':
                if d_val < val_range[1]:
                    filtered.append(d)
            elif op == '>':
                if d_val > val_range[1]:
                    filtered.append(d)
            elif op == '<=':
                if d_val <= val_range[1]:
                    filtered.append(d)
            elif op == '>=':
                if d_val >= val_range[1]:
                    filtered.append(d)
        return filtered
    else:
        # Exact number match
        for d in data:
            if int(d[key]) == val_range[1]:
                filtered.append(d)
        return filtered


def generate_patches_from_kwargs(operation, **kwargs):
    """Given a set of parameters, returns a list with the
    valid patch values.

    :param string operation: The operation to perform.
    :param list kwargs: Dict of parameters.

    :returns: A list with the right patch values.
    """
    patches = []
    for k, v in kwargs.items():
        patch = {'op': operation, 'value': v, 'path': f'/{k}'}
        patches.append(patch)
    return sorted(patches)


class FileSegment:
    """File-like object to pass to requests."""

    def __init__(self, filename, offset, length):
        self.filename = filename
        self.offset = offset
        self.length = length
        self.pos = 0
        self._file = open(filename, 'rb')
        self.seek(0)

    def tell(self):
        return self._file.tell() - self.offset

    def seek(self, offset, whence=0):
        if whence == 0:
            self._file.seek(self.offset + offset, whence)
        elif whence == 1:
            self._file.seek(offset, whence)
        elif whence == 2:
            self._file.seek(self.offset + self.length - offset, 0)

    def read(self, size=-1):
        remaining = self.length - self.pos
        if remaining <= 0:
            return b''

        to_read = remaining if size < 0 else min(size, remaining)
        chunk = self._file.read(to_read)
        self.pos += len(chunk)

        return chunk

    def reset(self):
        self._file.seek(self.offset, 0)


def _format_uuid_string(string):
    return (
        string.replace('urn:', '')
        .replace('uuid:', '')
        .strip('{}')
        .replace('-', '')
        .lower()
    )


def _is_uuid_like(val):
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionchanged:: 1.1.1
       Support non-lowercase UUIDs.
    """
    try:
        return str(uuid.UUID(val)).replace('-', '') == _format_uuid_string(val)
    except (TypeError, ValueError, AttributeError):
        return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import secret_acl as _secret_acl
from openstack.tests.functional import base


class TestSecretACL(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('key-manager')

        self.secret = self.operator_cloud.key_manager.create_secret(
            name=self.getUniqueString('secret'),
            payload='functional-secret-acl-payload',
            payload_content_type='text/plain',
            secret_type='opaque',
        )
        self.secret_id = self.secret.secret_id
        self.assertIsNotNone(self.secret_id)

        self.addCleanup(
            self.operator_cloud.key_manager.delete_secret,
            self.secret_id,
            ignore_missing=True,
        )
        self.addCleanup(
            self.operator_cloud.key_manager.delete_secret_acl,
            self.secret_id,
            ignore_missing=True,
        )

    def test_secret_acl(self):
        key_manager = self.operator_cloud.key_manager
        user_id = self.operator_cloud.current_user_id

        acl = key_manager.set_secret_acl(
            self.secret_id,
            read={'users': [user_id], 'project-access': False},
        )
        self.assertIsInstance(acl, _secret_acl.SecretACL)
        self.assertIsNotNone(acl.acl_ref)

        acl = key_manager.get_secret_acl(self.secret_id)
        self.assertIsInstance(acl, _secret_acl.SecretACL)
        self.assertFalse(acl.read['project-access'])
        self.assertIn(user_id, acl.read.get('users', []))

        acl = key_manager.update_secret_acl(
            self.secret_id, read={'project-access': True}
        )
        self.assertIsInstance(acl, _secret_acl.SecretACL)
        self.assertIsNotNone(acl.acl_ref)

        acl = key_manager.get_secret_acl(self.secret_id)
        self.assertTrue(acl.read['project-access'])

        self.assertIsNone(key_manager.delete_secret_acl(self.secret_id))

        acl = key_manager.get_secret_acl(self.secret_id)
        self.assertIsInstance(acl, _secret_acl.SecretACL)
        self.assertTrue(acl.read['project-access'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import network as _network
from openstack.network.v2 import port as _port
from openstack.network.v2 import tap_mirror as _tap_mirror
from openstack.tests.functional import base


class TestTapMirror(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        if not self.user_cloud.network.find_extension("tap-mirror"):
            self.skipTest("Neutron Tap Mirror Extension disabled")

        self.TAP_M_NAME = 'my_tap_mirror' + self.getUniqueString()
        net = self.user_cloud.network.create_network()
        assert isinstance(net, _network.Network)
        self.MIRROR_NET_ID = net.id

        port = self.user_cloud.network.create_port(
            network_id=self.MIRROR_NET_ID
        )
        assert isinstance(port, _port.Port)
        self.MIRROR_PORT_ID = port.id

        self.REMOTE_IP = '193.10.10.2'
        self.MIRROR_TYPE = 'erspanv1'

        tap_mirror = self.user_cloud.network.create_tap_mirror(
            name=self.TAP_M_NAME,
            port_id=self.MIRROR_PORT_ID,
            remote_ip=self.REMOTE_IP,
            mirror_type=self.MIRROR_TYPE,
            directions={'IN': 99},
        )
        assert isinstance(tap_mirror, _tap_mirror.TapMirror)
        self.TAP_MIRROR = tap_mirror

    def tearDown(self):
        sot = self.user_cloud.network.delete_tap_mirror(
            self.TAP_MIRROR.id, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_port(self.MIRROR_PORT_ID)
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(self.MIRROR_NET_ID)
        self.assertIsNone(sot)

        super().tearDown()

    def test_find_tap_mirror(self):
        sot = self.user_cloud.network.find_tap_mirror(
            self.TAP_MIRROR.name, ignore_missing=False
        )
        self.assertEqual(self.MIRROR_PORT_ID, sot.port_id)
        self.assertEqual(self.TAP_M_NAME, sot.name)

    def test_get_tap_mirror(self):
        sot = self.user_cloud.network.get_tap_mirror(self.TAP_MIRROR.id)
        self.assertEqual(self.MIRROR_PORT_ID, sot.port_id)
        self.assertEqual(self.TAP_M_NAME, sot.name)

    def test_list_tap_mirrors(self):
        tap_mirror_ids = [
            tm.id for tm in self.user_cloud.network.tap_mirrors()
        ]
        self.assertIn(self.TAP_MIRROR.id, tap_mirror_ids)

    def test_update_tap_mirror(self):
        description = 'My Tap Mirror'
        sot = self.user_cloud.network.update_tap_mirror(
            self.TAP_MIRROR.id, description=description
        )
        self.assertEqual(description, sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import quota_set
from openstack import resource


class QuotaSet(quota_set.QuotaSet):
    resource_key = "quota_set"
    resources_key = "quota_sets"
    base_path = "/quota-sets/%(project_id)s"

    # capabilities
    allow_create = False
    allow_list = False

    #: Properties
    #: The number of gigabytes allowed for each project.
    gigabytes = resource.Body("gigabytes", type=int)
    #: The number of snapshots allowed for each project.
    snapshots = resource.Body("snapshots", type=int)
    #: The number of shares allowed for each project.
    shares = resource.Body("shares", type=int)
    #: The number of gigabytes for the snapshots allowed for each project.
    snapshot_gigabytes = resource.Body("snapshot_gigabytes", type=int)
    #: The number of share groups allowed for each project or user.
    share_groups = resource.Body("share_groups", type=int)
    #: The number of share group snapshots allowed for each project or user.
    share_group_snapshots = resource.Body("share_group_snapshots", type=int)
    #: The number of share networks allowed for each project.
    share_networks = resource.Body("share_networks", type=int)
    #: The number of share replicas allowed for each project.
    share_replicas = resource.Body("share_replicas", type=int)
    #: The number of gigabytes for the share replicas allowed for each project.
    replica_gigabytes = resource.Body("replica_gigabytes", type=int)
    #: The number of gigabytes per share allowed for each project
    per_share_gigabytes = resource.Body("per_share_gigabytes", type=int)
    #: The number of backups allowed for each project.
    backups = resource.Body("backups", type=int)
    #: The number of gigabytes for the backups allowed for each project.
    backup_gigabytes = resource.Body("backup_gigabytes", type=int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import MutableMapping
from typing import Any, Self

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource


class AcceleratorRequest(resource.Resource):
    resource_key = 'arq'
    resources_key = 'arqs'
    base_path = '/accelerator_requests'

    create_opts = resource.CreateOpts(request_key=None)

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True
    #: Allow patch operation for binding.
    allow_patch = True

    #: The device address associated with this ARQ (if any)
    attach_handle_info = resource.Body('attach_handle_info')
    #: The type of attach handle (e.g. PCI, mdev...)
    attach_handle_type = resource.Body('attach_handle_type')
    #: The name of the device profile
    device_profile_name = resource.Body('device_profile_name')
    #: The id of the device profile group
    device_profile_group_id = resource.Body('device_profile_group_id')
    #: The UUID of the bound device RP (if any)
    device_rp_uuid = resource.Body('device_rp_uuid')
    #: The host name to which ARQ is bound. (if any)
    hostname = resource.Body('hostname')
    #: The UUID of the instance associated with this ARQ (if any)
    instance_uuid = resource.Body('instance_uuid')
    #: The state of the ARQ
    state = resource.Body('state')
    #: The UUID of the ARQ
    uuid = resource.Body('uuid', alternate_id=True)

    def _convert_patch(  # type: ignore[override]
        self, patch: list[dict[str, Any]]
    ) -> dict[str, Any]:
        # This overrides the default behavior of _convert_patch because
        # the PATCH method consumes JSON, its key is the ARQ uuid
        # and its value is an ordinary JSON patch. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api
        converted = super()._convert_patch(patch)
        return {self.id: converted}

    def patch(
        self,
        session: adapter.Adapter,
        patch: list[dict[str, Any]] | None = None,
        prepend_key: bool = True,
        has_body: bool = True,
        retry_on_conflict: bool | None = None,
        base_path: str | None = None,
        *,
        microversion: str | None = None,
    ) -> Self:
        # This overrides the default behavior of patch because
        # the PATCH method consumes a dict rather than a list. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api

        # The id cannot be dirty for an commit
        self._body._dirty.discard("id")

        # Only try to update if we actually have anything to commit.
        if not patch and not self.requires_commit:
            return self

        if not self.allow_patch:
            raise exceptions.MethodNotSupported(self, "patch")

        request = self._prepare_request(
            prepend_key=prepend_key, base_path=base_path, patch=True
        )
        microversion = self._get_microversion(session)
        if patch:
            request.body = self._convert_patch(patch)

        return self._commit(
            session,
            request,
            'PATCH',
            microversion,
            has_body=has_body,
            retry_on_conflict=retry_on_conflict,
        )

    def _consume_attrs(
        self,
        mapping: MutableMapping[str, Any],
        attrs: MutableMapping[str, Any],
    ) -> dict[str, Any]:
        # This overrides the default behavior of _consume_attrs because
        # cyborg api returns an ARQ as list. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api
        if isinstance(self, AcceleratorRequest):
            if self.resources_key in attrs:
                attrs = attrs[self.resources_key][0]
        return super()._consume_attrs(mapping, attrs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.tests.functional import base


class TestAvailabilityZone(base.BaseFunctionalTest):
    def test_list(self):
        availability_zones = list(self.user_cloud.network.availability_zones())
        if len(availability_zones) > 0:
            for az in availability_zones:
                self.assertIsInstance(az.name, str)
                self.assertIsInstance(az.resource, str)
                self.assertIsInstance(az.state, str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import exceptions
from openstack import resource
from openstack import utils


class LoadBalancer(resource.Resource, tag.TagMixin):
    resource_key = 'loadbalancer'
    resources_key = 'loadbalancers'
    base_path = '/lbaas/loadbalancers'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'flavor_id',
        'name',
        'project_id',
        'provider',
        'vip_address',
        'vip_network_id',
        'vip_port_id',
        'vip_subnet_id',
        'vip_qos_policy_id',
        'provisioning_status',
        'operating_status',
        'availability_zone',
        is_admin_state_up='admin_state_up',
        **tag.TagMixin._tag_query_parameters,
    )

    # Properties
    #: The administrative state of the load balancer *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: Name of the target Octavia availability zone
    availability_zone = resource.Body('availability_zone')
    #: Timestamp when the load balancer was created
    created_at = resource.Body('created_at')
    #: The load balancer description
    description = resource.Body('description')
    #: The load balancer flavor ID
    flavor_id = resource.Body('flavor_id')
    #: List of listeners associated with this load balancer
    listeners = resource.Body('listeners', type=list)
    #: The load balancer name
    name = resource.Body('name')
    #: Operating status of the load balancer
    operating_status = resource.Body('operating_status')
    #: List of pools associated with this load balancer
    pools = resource.Body('pools', type=list)
    #: The ID of the project this load balancer is associated with.
    project_id = resource.Body('project_id')
    #: Provider name for the load balancer.
    provider = resource.Body('provider')
    #: The provisioning status of this load balancer
    provisioning_status = resource.Body('provisioning_status')
    #: Timestamp when the load balancer was last updated
    updated_at = resource.Body('updated_at')
    #: VIP address of load balancer
    vip_address = resource.Body('vip_address')
    #: VIP netowrk ID
    vip_network_id = resource.Body('vip_network_id')
    #: VIP port ID
    vip_port_id = resource.Body('vip_port_id')
    #: VIP subnet ID
    vip_subnet_id = resource.Body('vip_subnet_id')
    # VIP qos policy id
    vip_qos_policy_id = resource.Body('vip_qos_policy_id')
    #: List of security group IDs for the VIP port
    vip_sg_ids = resource.Body('vip_sg_ids', type=list)
    #: Additional VIPs
    additional_vips = resource.Body('additional_vips', type=list)

    def delete(
        self,
        session: adapter.Adapter,
        error_message: str | None = None,
        **kwargs: Any,
    ) -> Self:
        request = self._prepare_request()
        params = {}
        if (
            hasattr(self, 'cascade')
            and isinstance(self.cascade, bool)
            and self.cascade
        ):
            params['cascade'] = True
        response = session.delete(request.url, params=params)

        self._translate_response(
            response, has_body=False, error_message=error_message
        )
        return self

    def failover(self, session: resource.AdapterT) -> None:
        """Failover load balancer.

        :param session: The session to use for making this request.
        :returns: None
        """
        session = self._get_session(session)
        version = self._get_microversion(session)
        request = self._prepare_request(requires_id=True)
        request.url = utils.urljoin(request.url, 'failover')

        response = session.put(
            request.url,
            headers=request.headers,
            microversion=version,
        )

        msg = f"Failed to failover load balancer {self.id}"
        exceptions.raise_from_response(response, error_message=msg)


class LoadBalancerStats(resource.Resource):
    resource_key = 'stats'
    base_path = '/lbaas/loadbalancers/%(lb_id)s/stats'

    # capabilities
    allow_create = False
    allow_fetch = True
    allow_commit = False
    allow_delete = False
    allow_list = False

    # Properties
    #: The ID of the load balancer.
    lb_id = resource.URI('lb_id')
    #: The currently active connections.
    active_connections = resource.Body('active_connections', type=int)
    #: The total bytes received.
    bytes_in = resource.Body('bytes_in', type=int)
    #: The total bytes sent.
    bytes_out = resource.Body('bytes_out', type=int)
    #: The total requests that were unable to be fulfilled.
    request_errors = resource.Body('request_errors', type=int)
    #: The total connections handled.
    total_connections = resource.Body('total_connections', type=int)


# TODO(stephenfin): Delete this: it's useless
class LoadBalancerFailover(resource.Resource):
    base_path = '/lbaas/loadbalancers/%(lb_id)s/failover'

    # capabilities
    allow_create = False
    allow_fetch = False
    allow_commit = True
    allow_delete = False
    allow_list = False
    allow_empty_commit = True

    requires_id = False

    # Properties
    #: The ID of the load balancer.
    lb_id = resource.URI('lb_id')

    # The default _update code path also has no
    # way to pass has_body into this function, so overriding the method here.
    def commit(
        self,
        session: adapter.Adapter,
        prepend_key: bool = True,
        has_body: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        return super().commit(session, prepend_key, has_body, *args, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self, cast

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class Transfer(resource.Resource):
    resource_key = "transfer"
    resources_key = "transfers"
    base_path = "/volume-transfers"

    # capabilities
    allow_create = True
    allow_delete = True
    allow_fetch = True
    allow_list = True

    # Properties
    #: UUID of the transfer.
    id = resource.Body("id")
    #: The date and time when the resource was created.
    created_at = resource.Body("created_at")
    #: Name of the volume to transfer.
    name = resource.Body("name")
    #: ID of the volume to transfer.
    volume_id = resource.Body("volume_id")
    #: Auth key for the transfer.
    auth_key = resource.Body("auth_key")
    #: A list of links associated with this volume. *Type: list*
    links = resource.Body("links")
    #: Whether to transfer snapshots or not
    no_snapshots = resource.Body("no_snapshots")

    _max_microversion = "3.55"

    # TODO(stephenfin): Migrate to _transform_create_request. The body cleanup
    # (no_snapshots removal) is doable in the hook, but the base_path switch
    # from /volume-transfers to /os-volume-transfer requires URL manipulation
    # that is fragile and best handled by a dedicated hook or CreateOpts field.
    def create(
        self,
        session: adapter.Adapter,
        prepend_key: bool = True,
        base_path: str | None = None,
        *,
        microversion: str | None = None,
        **params: Any,
    ) -> Self:
        # With MV 3.55 we introduced new API for volume transfer
        # (/volume-transfers). Prior to that (MV < 3.55), we use
        # the old API (/os-volume-transfer)
        if not utils.supports_microversion(session, '3.55'):
            base_path = '/os-volume-transfer'
            # With MV 3.55, we also introduce the ability to transfer
            # snapshot along with the volume. If MV < 3.55, we should
            # not send 'no_snapshots' parameter in the request.
            if 'no_snapshots' in params:
                params.pop('no_snapshots')

        return super().create(
            session,
            prepend_key=prepend_key,
            base_path=base_path,
            microversion=microversion,
            **params,
        )

    def fetch(
        self,
        session: adapter.Adapter,
        requires_id: bool = True,
        base_path: str | None = None,
        error_message: str | None = None,
        skip_cache: bool = False,
        *,
        resource_response_key: str | None = None,
        microversion: str | None = None,
        **params: Any,
    ) -> Self:
        """Get volume transfer.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param boolean requires_id: A boolean indicating whether resource ID
            should be part of the requested URI.
        :param str base_path: Base part of the URI for fetching resources, if
            different from :data:`~openstack.resource.Resource.base_path`.
        :param str error_message: An Error message to be returned if
            requested object does not exist.
        :param bool skip_cache: A boolean indicating whether optional API
            cache should be skipped for this invocation.
        :param str resource_response_key: Overrides the usage of
            self.resource_key when processing the response body.
        :param str microversion: API version to override the negotiated one.
        :param dict params: Additional parameters that can be consumed.
        :return: This :class:`Resource` instance.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
            :data:`Resource.allow_fetch` is not set to ``True``.
        :raises: :exc:`~openstack.exceptions.NotFoundException` if
            the resource was not found.
        """

        if not utils.supports_microversion(session, '3.55'):
            base_path = '/os-volume-transfer'

        return super().fetch(
            session,
            requires_id=requires_id,
            base_path=base_path,
            error_message=error_message,
            skip_cache=skip_cache,
            resource_response_key=resource_response_key,
            microversion=microversion,
            **params,
        )

    def delete(
        self,
        session: adapter.Adapter,
        error_message: str | None = None,
        *,
        microversion: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Delete a volume transfer.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param str microversion: API version to override the negotiated one.
        :param dict kwargs: Parameters that will be passed to
            _prepare_request()

        :return: This :class:`Resource` instance.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
            :data:`Resource.allow_commit` is not set to ``True``.
        :raises: :exc:`~openstack.exceptions.NotFoundException` if
            the resource was not found.
        """

        if not utils.supports_microversion(session, '3.55'):
            kwargs['base_path'] = '/os-volume-transfer'
        return super().delete(
            session,
            error_message=error_message,
            microversion=microversion,
            **kwargs,
        )

    def accept(
        self, session: adapter.Adapter, *, auth_key: str | None = None
    ) -> Self:
        """Accept a volume transfer.

        :param session: The session to use for making this request.
        :param auth_key: The authentication key for the volume transfer.

        :return: This :class:`Transfer` instance.
        """
        body = {'accept': {'auth_key': auth_key}}

        path = self.base_path
        if not utils.supports_microversion(session, '3.55'):
            path = '/os-volume-transfer'

        url = utils.urljoin(path, self.id, 'accept')
        microversion = self._get_microversion(session)
        resp = session.post(
            url,
            json=body,
            microversion=microversion,
        )
        exceptions.raise_from_response(resp)

        transfer = Transfer()
        transfer._translate_response(resp)
        return cast(Self, transfer)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_transfer
from openstack.tests.unit import base

EXAMPLE = [
    {
        "id": "02a948b4-671b-4c62-b13a-18d613cb4576",
        "resource_type": "share",
        "resource_id": "0fe7cf64-b879-4902-9d86-f80aeff12b06",
        "name": "transfer2",
        "links": [
            {
                "rel": "self",
                "href": "http://127.0.0.1/share/v2/share-transfer"
                "/02a948b4-671b-4c62-b13a-18d613cb4576",
            },
            {
                "rel": "bookmark",
                "href": "http://127.0.0.1/share/share-transfer/02a948b4"
                "-671b-4c62-b13a-18d613cb4576",
            },
        ],
    },
    {
        "id": "a10209ff-b55d-4fed-9f63-abea53b6f107",
        "resource_type": "share",
        "resource_id": "29476819-28a9-4b1a-a21d-3b2d203025a0",
        "name": "transfer1",
        "links": [
            {
                "rel": "self",
                "href": "http://127.0.0.1/share/v2/share-transfer"
                "/a10209ff-b55d-4fed-9f63-abea53b6f107",
            },
            {
                "rel": "bookmark",
                "href": "http://127.0.0.1/share/share-transfer/a10209ff"
                "-b55d-4fed-9f63-abea53b6f107",
            },
        ],
    },
]


class TestShareTransfer(base.TestCase):
    def test_basic(self):
        transfers = share_transfer.ShareTransfer()
        self.assertTrue(transfers.allow_list)
        self.assertTrue(transfers.allow_fetch)
        self.assertTrue(transfers.allow_create)
        self.assertFalse(transfers.allow_commit)
        self.assertTrue(transfers.allow_delete)

    def test_get_share_transfer_details(self):
        transfer = share_transfer.ShareTransfer.existing(**EXAMPLE[0])
        self.assertEqual(EXAMPLE[0]['id'], transfer.id)
        self.assertEqual(EXAMPLE[0]['resource_type'], transfer.resource_type)
        self.assertEqual(EXAMPLE[0]['resource_id'], transfer.resource_id)
        self.assertEqual(EXAMPLE[0]['name'], transfer.name)
        self.assertEqual(EXAMPLE[0]['links'], transfer.links)



# Copyright 2026 Openinfra Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.compute.v2 import server_external_event
from openstack.tests.unit import base


EXAMPLE = {
    'name': 'testevent',
    'server_uuid': 'SERVER_UUID',
    'status': 'completed',
    'tag': 'MYTAG',
}


class TestServerExternalEvent(base.TestCase):
    def test_basic(self):
        sot = server_external_event.ServerExternalEvent()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('events', sot.resources_key)
        self.assertEqual('/os-server-external-events', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = server_external_event.ServerExternalEvent(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['server_uuid'], sot.server_uuid)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['tag'], sot.tag)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests.functional.baremetal.v1 import base


class TestBareMetalDeployTemplate(base.BaseBaremetalTest):
    min_microversion = '1.55'

    def test_baremetal_deploy_create_get_delete(self):
        steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {
                    "settings": [{"name": "LogicalProc", "value": "Enabled"}]
                },
                "priority": 150,
            }
        ]
        deploy_template = self.create_deploy_template(
            name='CUSTOM_DEPLOY_TEMPLATE', steps=steps
        )
        loaded = self.system_admin_cloud.baremetal.get_deploy_template(
            deploy_template.id
        )
        self.assertEqual(loaded.id, deploy_template.id)
        self.system_admin_cloud.baremetal.delete_deploy_template(
            deploy_template, ignore_missing=False
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.system_admin_cloud.baremetal.get_deploy_template,
            deploy_template.id,
        )

    def test_baremetal_deploy_template_list(self):
        steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {
                    "settings": [{"name": "LogicalProc", "value": "Enabled"}]
                },
                "priority": 150,
            }
        ]

        deploy_template1 = self.create_deploy_template(
            name='CUSTOM_DEPLOY_TEMPLATE1', steps=steps
        )
        deploy_template2 = self.create_deploy_template(
            name='CUSTOM_DEPLOY_TEMPLATE2', steps=steps
        )
        deploy_templates = self.system_admin_cloud.baremetal.deploy_templates()
        ids = [template.id for template in deploy_templates]
        self.assertIn(deploy_template1.id, ids)
        self.assertIn(deploy_template2.id, ids)

        deploy_templates_with_details = (
            self.system_admin_cloud.baremetal.deploy_templates(details=True)
        )
        for dp in deploy_templates_with_details:
            self.assertIsNotNone(dp.id)
            self.assertIsNotNone(dp.name)

        deploy_tempalte_with_fields = (
            self.system_admin_cloud.baremetal.deploy_templates(fields=['uuid'])
        )
        for dp in deploy_tempalte_with_fields:
            self.assertIsNotNone(dp.id)
            self.assertIsNone(dp.name)

    def test_baremetal_deploy_list_update_delete(self):
        steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {
                    "settings": [{"name": "LogicalProc", "value": "Enabled"}]
                },
                "priority": 150,
            }
        ]
        deploy_template = self.create_deploy_template(
            name='CUSTOM_DEPLOY_TEMPLATE4', steps=steps
        )
        self.assertFalse(deploy_template.extra)
        deploy_template.extra = {'answer': 42}

        deploy_template = (
            self.system_admin_cloud.baremetal.update_deploy_template(
                deploy_template
            )
        )
        self.assertEqual({'answer': 42}, deploy_template.extra)

        deploy_template = (
            self.system_admin_cloud.baremetal.get_deploy_template(
                deploy_template.id
            )
        )

        self.system_admin_cloud.baremetal.delete_deploy_template(
            deploy_template.id, ignore_missing=False
        )

    def test_baremetal_deploy_update(self):
        steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {
                    "settings": [{"name": "LogicalProc", "value": "Enabled"}]
                },
                "priority": 150,
            }
        ]
        deploy_template = self.create_deploy_template(
            name='CUSTOM_DEPLOY_TEMPLATE4', steps=steps
        )
        deploy_template.extra = {'answer': 42}

        deploy_template = (
            self.system_admin_cloud.baremetal.update_deploy_template(
                deploy_template
            )
        )
        self.assertEqual({'answer': 42}, deploy_template.extra)

        deploy_template = (
            self.system_admin_cloud.baremetal.get_deploy_template(
                deploy_template.id
            )
        )
        self.assertEqual({'answer': 42}, deploy_template.extra)

    def test_deploy_template_patch(self):
        name = "CUSTOM_HYPERTHREADING_ON"
        steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {
                    "settings": [{"name": "LogicalProc", "value": "Enabled"}]
                },
                "priority": 150,
            }
        ]
        deploy_template = self.create_deploy_template(name=name, steps=steps)
        deploy_template = (
            self.system_admin_cloud.baremetal.patch_deploy_template(
                deploy_template,
                [{'path': '/extra/answer', 'op': 'add', 'value': 42}],
            )
        )
        self.assertEqual({'answer': 42}, deploy_template.extra)
        self.assertEqual(name, deploy_template.name)

        deploy_template = (
            self.system_admin_cloud.baremetal.get_deploy_template(
                deploy_template.id
            )
        )
        self.assertEqual({'answer': 42}, deploy_template.extra)

    def test_deploy_template_negative_non_existing(self):
        uuid = "bbb45f41-d4bc-4307-8d1d-32f95ce1e920"
        self.assertRaises(
            exceptions.NotFoundException,
            self.system_admin_cloud.baremetal.get_deploy_template,
            uuid,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.system_admin_cloud.baremetal.delete_deploy_template,
            uuid,
            ignore_missing=False,
        )
        self.assertIsNone(
            self.system_admin_cloud.baremetal.delete_deploy_template(uuid)
        )



#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstack.network.v2 import local_ip_association
from openstack.tests.unit import base

EXAMPLE = {
    'local_ip_id': '0',
    'local_ip_address': '127.0.0.1',
    'fixed_port_id': '1',
    'fixed_ip': '127.0.0.2',
    'host': '2',
}


class TestLocalIP(base.TestCase):
    def test_basic(self):
        sot = local_ip_association.LocalIPAssociation()
        self.assertEqual('port_association', sot.resource_key)
        self.assertEqual('port_associations', sot.resources_key)
        self.assertEqual(
            '/local_ips/%(local_ip_id)s/port_associations', sot.base_path
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'fixed_port_id': 'fixed_port_id',
                'fixed_ip': 'fixed_ip',
                'host': 'host',
                'limit': 'limit',
                'marker': 'marker',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = local_ip_association.LocalIPAssociation(**EXAMPLE)
        self.assertEqual(EXAMPLE['local_ip_id'], sot.local_ip_id)
        self.assertEqual(EXAMPLE['local_ip_address'], sot.local_ip_address)
        self.assertEqual(EXAMPLE['fixed_port_id'], sot.fixed_port_id)
        self.assertEqual(EXAMPLE['fixed_ip'], sot.fixed_ip)
        self.assertEqual(EXAMPLE['host'], sot.host)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import io
import operator
import os
import tempfile
from unittest import mock

from keystoneauth1 import adapter
import requests

from openstack import _log
from openstack import exceptions
from openstack.image.v2 import image
from openstack.tests.unit import base

IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'id': IDENTIFIER,
    'checksum': '1',
    'container_format': '2',
    'created_at': '2015-03-09T12:14:57.233772',
    'data': 'This is not an image',
    'disk_format': '4',
    'min_disk': 5,
    'name': '6',
    'owner': '7',
    'properties': {
        'a': 'z',
        'b': 'y',
    },
    'protected': False,
    'status': '8',
    'tags': ['g', 'h', 'i'],
    'updated_at': '2015-03-09T12:15:57.233772',
    'os_hash_algo': 'sha512',
    'os_hash_value': '073b4523583784fbe01daff81eba092a262ec3',
    'os_hidden': False,
    'virtual_size': '10',
    'visibility': '11',
    'location': '12',
    'size': 13,
    'store': '14',
    'file': '15',
    'locations': ['15', '16'],
    'direct_url': '17',
    'url': '20',
    'metadata': {'21': '22'},
    'architecture': '23',
    'hypervisor_type': '24',
    'instance_type_rxtx_factor': 25.1,
    'instance_uuid': '26',
    'img_config_drive': '27',
    'kernel_id': '28',
    'os_distro': '29',
    'os_version': '30',
    'os_secure_boot': '31',
    'ramdisk_id': '32',
    'vm_mode': '33',
    'hw_cpu_sockets': 34,
    'hw_cpu_cores': 35,
    'hw_cpu_threads': 36,
    'hw_disk_bus': '37',
    'hw_rng_model': '38',
    'hw_machine_type': '39',
    'hw_scsi_model': '40',
    'hw_serial_port_count': 41,
    'hw_video_model': '42',
    'hw_video_ram': 43,
    'hw_watchdog_action': '44',
    'os_command_line': '45',
    'hw_vif_model': '46',
    'hw_vif_multiqueue_enabled': True,
    'hw_boot_menu': True,
    'vmware_adaptertype': '47',
    'vmware_ostype': '48',
    'auto_disk_config': True,
    'os_type': '49',
    'os_admin_user': 'ubuntu',
    'hw_qemu_guest_agent': 'yes',
    'os_require_quiesce': True,
}


def calculate_md5_checksum(data):
    checksum = hashlib.md5(usedforsecurity=False)
    for chunk in data:
        checksum.update(chunk)
    return checksum.hexdigest()


class FakeResponse:
    def __init__(self, response, status_code=200, headers=None, reason=None):
        self.body = response
        self.content = response
        self.text = response
        self.status_code = status_code
        headers = headers if headers else {'content-type': 'application/json'}
        self.headers = requests.structures.CaseInsensitiveDict(headers)
        if reason:
            self.reason = reason
        # for the sake of "list" response faking
        self.links = []

    def json(self):
        return self.body


class TestImage(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess.put = mock.Mock(return_value=FakeResponse({}))
        self.sess.delete = mock.Mock(return_value=FakeResponse({}))
        self.sess.get = mock.Mock(return_value=FakeResponse({}))
        self.sess.default_microversion = None
        self.sess.retriable_status_codes = None
        self.sess.log = _log.setup_logging('openstack')

    def test_basic(self):
        sot = image.Image()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('images', sot.resources_key)
        self.assertEqual('/images', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'created_at': 'created_at',
                'id': 'id',
                'is_hidden': 'os_hidden',
                'limit': 'limit',
                'marker': 'marker',
                'member_status': 'member_status',
                'name': 'name',
                'owner': 'owner',
                'protected': 'protected',
                'size_max': 'size_max',
                'size_min': 'size_min',
                'sort': 'sort',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
                'status': 'status',
                'tag': 'tag',
                'updated_at': 'updated_at',
                'visibility': 'visibility',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = image.Image(**EXAMPLE)
        self.assertEqual(IDENTIFIER, sot.id)
        self.assertEqual(EXAMPLE['checksum'], sot.checksum)
        self.assertEqual(EXAMPLE['container_format'], sot.container_format)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['disk_format'], sot.disk_format)
        self.assertEqual(EXAMPLE['min_disk'], sot.min_disk)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['owner'], sot.owner)
        self.assertEqual(EXAMPLE['owner'], sot.owner_id)
        self.assertEqual(EXAMPLE['properties'], sot.properties)
        self.assertFalse(sot.is_protected)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['tags'], sot.tags)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['os_hash_algo'], sot.hash_algo)
        self.assertEqual(EXAMPLE['os_hash_value'], sot.hash_value)
        self.assertEqual(EXAMPLE['os_hidden'], sot.is_hidden)
        self.assertEqual(EXAMPLE['virtual_size'], sot.virtual_size)
        self.assertEqual(EXAMPLE['visibility'], sot.visibility)
        self.assertEqual(EXAMPLE['size'], sot.size)
        self.assertEqual(EXAMPLE['store'], sot.store)
        self.assertEqual(EXAMPLE['file'], sot.file)
        self.assertEqual(EXAMPLE['locations'], sot.locations)
        self.assertEqual(EXAMPLE['direct_url'], sot.direct_url)
        self.assertEqual(EXAMPLE['url'], sot.url)
        self.assertEqual(EXAMPLE['metadata'], sot.metadata)
        self.assertEqual(EXAMPLE['architecture'], sot.architecture)
        self.assertEqual(EXAMPLE['hypervisor_type'], sot.hypervisor_type)
        self.assertEqual(
            EXAMPLE['instance_type_rxtx_factor'], sot.instance_type_rxtx_factor
        )
        self.assertEqual(EXAMPLE['instance_uuid'], sot.instance_uuid)
        self.assertEqual(EXAMPLE['img_config_drive'], sot.needs_config_drive)
        self.assertEqual(EXAMPLE['kernel_id'], sot.kernel_id)
        self.assertEqual(EXAMPLE['os_distro'], sot.os_distro)
        self.assertEqual(EXAMPLE['os_version'], sot.os_version)
        self.assertEqual(EXAMPLE['os_secure_boot'], sot.needs_secure_boot)
        self.assertEqual(EXAMPLE['ramdisk_id'], sot.ramdisk_id)
        self.assertEqual(EXAMPLE['vm_mode'], sot.vm_mode)
        self.assertEqual(EXAMPLE['hw_cpu_sockets'], sot.hw_cpu_sockets)
        self.assertEqual(EXAMPLE['hw_cpu_cores'], sot.hw_cpu_cores)
        self.assertEqual(EXAMPLE['hw_cpu_threads'], sot.hw_cpu_threads)
        self.assertEqual(EXAMPLE['hw_disk_bus'], sot.hw_disk_bus)
        self.assertEqual(EXAMPLE['hw_rng_model'], sot.hw_rng_model)
        self.assertEqual(EXAMPLE['hw_machine_type'], sot.hw_machine_type)
        self.assertEqual(EXAMPLE['hw_scsi_model'], sot.hw_scsi_model)
        self.assertEqual(
            EXAMPLE['hw_serial_port_count'], sot.hw_serial_port_count
        )
        self.assertEqual(EXAMPLE['hw_video_model'], sot.hw_video_model)
        self.assertEqual(EXAMPLE['hw_video_ram'], sot.hw_video_ram)
        self.assertEqual(EXAMPLE['hw_watchdog_action'], sot.hw_watchdog_action)
        self.assertEqual(EXAMPLE['os_command_line'], sot.os_command_line)
        self.assertEqual(EXAMPLE['hw_vif_model'], sot.hw_vif_model)
        self.assertEqual(
            EXAMPLE['hw_vif_multiqueue_enabled'],
            sot.is_hw_vif_multiqueue_enabled,
        )
        self.assertEqual(EXAMPLE['hw_boot_menu'], sot.is_hw_boot_menu_enabled)
        self.assertEqual(EXAMPLE['vmware_adaptertype'], sot.vmware_adaptertype)
        self.assertEqual(EXAMPLE['vmware_ostype'], sot.vmware_ostype)
        self.assertEqual(EXAMPLE['auto_disk_config'], sot.has_auto_disk_config)
        self.assertEqual(EXAMPLE['os_type'], sot.os_type)
        self.assertEqual(EXAMPLE['os_admin_user'], sot.os_admin_user)
        self.assertEqual(
            EXAMPLE['hw_qemu_guest_agent'], sot.hw_qemu_guest_agent
        )
        self.assertEqual(EXAMPLE['os_require_quiesce'], sot.os_require_quiesce)

    def test_deactivate(self):
        sot = image.Image(**EXAMPLE)
        self.assertIsNone(sot.deactivate(self.sess))
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/actions/deactivate',
        )

    def test_reactivate(self):
        sot = image.Image(**EXAMPLE)
        self.assertIsNone(sot.reactivate(self.sess))
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/actions/reactivate',
        )

    def test_add_tag(self):
        sot = image.Image(**EXAMPLE)
        tag = "lol"

        sot.add_tag(self.sess, tag)
        self.sess.put.assert_called_with(
            f'images/IDENTIFIER/tags/{tag}',
            microversion=None,
        )

    def test_remove_tag(self):
        sot = image.Image(**EXAMPLE)
        tag = "lol"

        sot.remove_tag(self.sess, tag)
        self.sess.delete.assert_called_with(
            f'images/IDENTIFIER/tags/{tag}',
            microversion=None,
        )

    def test_import_image(self):
        sot = image.Image(**EXAMPLE)
        json = {"method": {"name": "web-download", "uri": "such-a-good-uri"}}
        sot.import_image(self.sess, "web-download", uri="such-a-good-uri")
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/import', headers={}, json=json
        )

    def test_import_image_with_uri_not_web_download(self):
        sot = image.Image(**EXAMPLE)

        sot.import_image(self.sess, "glance-direct")
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/import',
            headers={},
            json={"method": {"name": "glance-direct"}},
        )

    def test_import_image_with_store(self):
        sot = image.Image(**EXAMPLE)
        json = {
            "method": {
                "name": "web-download",
                "uri": "such-a-good-uri",
            },
            "stores": ["ceph_1"],
        }
        store = mock.MagicMock()
        store.id = "ceph_1"
        sot.import_image(
            self.sess,
            "web-download",
            uri="such-a-good-uri",
            store=store,
        )
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/import',
            headers={'X-Image-Meta-Store': 'ceph_1'},
            json=json,
        )

    def test_import_image_with_stores(self):
        sot = image.Image(**EXAMPLE)
        json = {
            "method": {
                "name": "web-download",
                "uri": "such-a-good-uri",
            },
            "stores": ["ceph_1"],
        }
        store = mock.MagicMock()
        store.id = "ceph_1"
        sot.import_image(
            self.sess,
            "web-download",
            uri="such-a-good-uri",
            stores=[store],
        )
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/import',
            headers={},
            json=json,
        )

    def test_import_image_with_all_stores(self):
        sot = image.Image(**EXAMPLE)
        json = {
            "method": {
                "name": "web-download",
                "uri": "such-a-good-uri",
            },
            "all_stores": True,
        }
        sot.import_image(
            self.sess,
            "web-download",
            uri="such-a-good-uri",
            all_stores=True,
        )
        self.sess.post.assert_called_with(
            'images/IDENTIFIER/import',
            headers={},
            json=json,
        )

    def test_upload(self):
        sot = image.Image(**EXAMPLE)

        self.assertIsNotNone(sot.upload(self.sess))
        self.sess.put.assert_called_with(
            'images/IDENTIFIER/file',
            data=sot.data,
            headers={"Content-Type": "application/octet-stream", "Accept": ""},
        )

    def test_stage(self):
        sot = image.Image(**EXAMPLE)

        self.assertIsNotNone(sot.stage(self.sess))
        self.sess.put.assert_called_with(
            'images/IDENTIFIER/stage',
            data=sot.data,
            headers={"Content-Type": "application/octet-stream", "Accept": ""},
        )

    def test_stage_error(self):
        sot = image.Image(**EXAMPLE)

        self.sess.put.return_value = FakeResponse("dummy", status_code=400)
        self.assertRaises(exceptions.SDKException, sot.stage, self.sess)

    def test_download_checksum_match(self):
        expected_hash = hashlib.sha512(b"abc").hexdigest()
        example_with_hash = EXAMPLE.copy()
        example_with_hash['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_hash)

        resp1 = FakeResponse(example_with_hash)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        rv = sot.download(self.sess)
        self.sess.get.assert_has_calls(
            [
                mock.call(
                    'images/IDENTIFIER',
                    microversion=None,
                    params={},
                    skip_cache=False,
                ),
                mock.call('images/IDENTIFIER/file', stream=False),
            ]
        )

        self.assertEqual(rv, resp2)

    def test_download_checksum_mismatch(self):
        example_with_wrong_hash = EXAMPLE.copy()
        example_with_wrong_hash['os_hash_value'] = "wrong_hash_value"
        sot = image.Image(**example_with_wrong_hash)

        resp1 = FakeResponse(example_with_wrong_hash)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        self.assertRaises(exceptions.InvalidResponse, sot.download, self.sess)

    def test_download_md5_fallback(self):
        expected_md5 = hashlib.md5(b"abc", usedforsecurity=False).hexdigest()
        example_md5_only = EXAMPLE.copy()
        example_md5_only['os_hash_algo'] = None
        example_md5_only['os_hash_value'] = None
        example_md5_only['checksum'] = expected_md5
        sot = image.Image(**example_md5_only)

        resp1 = FakeResponse(example_md5_only)
        resp2 = FakeResponse(
            b"abc", headers={"Content-Type": "application/octet-stream"}
        )

        self.sess.get.side_effect = [resp1, resp2]

        rv = sot.download(self.sess)
        self.sess.get.assert_has_calls(
            [
                mock.call(
                    'images/IDENTIFIER',
                    microversion=None,
                    params={},
                    skip_cache=False,
                ),
                mock.call('images/IDENTIFIER/file', stream=False),
            ]
        )

        self.assertEqual(rv, resp2)

    def test_download_no_checksum_at_all2(self):
        # No hash available at all
        example_no_hash = EXAMPLE.copy()
        example_no_hash['os_hash_algo'] = None
        example_no_hash['os_hash_value'] = None
        example_no_hash['checksum'] = None
        sot = image.Image(**example_no_hash)

        resp1 = FakeResponse(example_no_hash)
        resp2 = FakeResponse(
            b"abc", headers={"Content-Type": "application/octet-stream"}
        )

        self.sess.get.side_effect = [resp1, resp2]

        with self.assertLogs(logger='openstack', level="WARNING") as log:
            rv = sot.download(self.sess)

            self.assertEqual(
                len(log.records), 1, "Too many warnings were logged"
            )
            self.assertEqual(
                "Unable to verify the integrity of image %s "
                "- no hash available",
                log.records[0].msg,
            )
            self.assertEqual((sot.id,), log.records[0].args)

        self.sess.get.assert_has_calls(
            [
                mock.call(
                    'images/IDENTIFIER',
                    microversion=None,
                    params={},
                    skip_cache=False,
                ),
                mock.call('images/IDENTIFIER/file', stream=False),
            ]
        )

        self.assertEqual(rv, resp2)

    def test_download_stream(self):
        expected_hash = hashlib.sha512(b"abc").hexdigest()
        example_with_hash = EXAMPLE.copy()
        example_with_hash['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_hash)

        resp1 = FakeResponse(example_with_hash)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        test_data = [
            {
                'name': 'Without store preferences',
                'store_preferences': None,
                'mock_call': 'images/IDENTIFIER/file',
            },
            {
                'name': 'With store preferences',
                'store_preferences': ['ceph', 's3'],
                'mock_call': 'images/IDENTIFIER/file?prefer=ceph,s3',
            },
        ]

        for data in test_data:
            with self.subTest(msg=data['name']):
                self.sess.get.side_effect = [resp1, resp2]
                rv = sot.download(
                    self.sess,
                    stream=True,
                    store_preferences=data['store_preferences'],
                )
                self.sess.get.assert_has_calls(
                    [
                        mock.call(
                            'images/IDENTIFIER',
                            microversion=None,
                            params={},
                            skip_cache=False,
                        ),
                        mock.call(data['mock_call'], stream=True),
                    ]
                )

                self.assertEqual(rv, resp2)
                self.assertIsNone(rv.headers.get('content-md5'))

    def test_image_download_output_fd(self):
        output_file = io.BytesIO()
        expected_hash = hashlib.sha512(b'0102').hexdigest()
        example_with_hash = EXAMPLE.copy()
        example_with_hash['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_hash)

        fetch_response = FakeResponse(example_with_hash)
        response = mock.Mock()
        response.status_code = 200
        response.iter_content.return_value = [b'01', b'02']
        response.headers = {}

        self.sess.get = mock.Mock(side_effect=[fetch_response, response])
        sot.download(self.sess, output=output_file)
        output_file.seek(0)
        self.assertEqual(b'0102', output_file.read())

    def test_image_download_output_file(self):
        expected_hash = hashlib.sha512(b'0102').hexdigest()
        example_with_hash = EXAMPLE.copy()
        example_with_hash['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_hash)

        fetch_response = FakeResponse(example_with_hash)
        response = mock.Mock()
        response.status_code = 200
        response.iter_content.return_value = [b'01', b'02']
        response.headers = {}

        self.sess.get = mock.Mock(side_effect=[fetch_response, response])

        output_file = tempfile.NamedTemporaryFile(delete=False)
        output_file.close()
        try:
            sot.download(self.sess, output=output_file.name)
            with open(output_file.name, 'rb') as fd:
                self.assertEqual(b'0102', fd.read())
        finally:
            os.unlink(output_file.name)

    def test_download_secure_hash_sha256(self):
        expected_hash = hashlib.sha256(b"abc").hexdigest()
        example_with_sha256 = EXAMPLE.copy()
        example_with_sha256['os_hash_algo'] = 'sha256'
        example_with_sha256['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_sha256)

        resp1 = FakeResponse(example_with_sha256)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        rv = sot.download(self.sess)
        self.assertEqual(rv, resp2)

    def test_download_secure_hash_sha384(self):
        expected_hash = hashlib.sha384(b"abc").hexdigest()
        example_with_sha384 = EXAMPLE.copy()
        example_with_sha384['os_hash_algo'] = 'sha384'
        example_with_sha384['os_hash_value'] = expected_hash
        sot = image.Image(**example_with_sha384)

        resp1 = FakeResponse(example_with_sha384)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        rv = sot.download(self.sess)
        self.assertEqual(rv, resp2)

    def test_download_content_md5_header_ignored(self):
        correct_hash = hashlib.sha512(b"abc").hexdigest()
        example_with_hash = EXAMPLE.copy()
        example_with_hash['os_hash_value'] = correct_hash
        sot = image.Image(**example_with_hash)

        resp1 = FakeResponse(example_with_hash)
        resp2 = FakeResponse(
            b"abc",
            headers={
                "Content-MD5": "wrong_header_hash_that_should_be_ignored",
                "Content-Type": "application/octet-stream",
            },
        )

        self.sess.get.side_effect = [resp1, resp2]

        # Succeeds since only metadata hash is used
        rv = sot.download(self.sess)
        self.assertEqual(rv, resp2)

    def test_download_secure_hash_mismatch_sha512(self):
        example_with_wrong_hash = EXAMPLE.copy()
        example_with_wrong_hash['os_hash_value'] = "wrong_sha512_hash"
        sot = image.Image(**example_with_wrong_hash)

        resp1 = FakeResponse(example_with_wrong_hash)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        self.assertRaises(exceptions.InvalidResponse, sot.download, self.sess)

    def test_download_md5_fallback_mismatch(self):
        example_md5_only = EXAMPLE.copy()
        example_md5_only['os_hash_algo'] = None
        example_md5_only['os_hash_value'] = None
        example_md5_only['checksum'] = "wrong_md5_checksum"
        sot = image.Image(**example_md5_only)

        resp1 = FakeResponse(example_md5_only)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        self.assertRaises(exceptions.InvalidResponse, sot.download, self.sess)

    def test_download_unsupported_hash_algo_raises(self):
        example_unsupported = EXAMPLE.copy()
        example_unsupported['os_hash_algo'] = 'unsupported_algo'
        example_unsupported['os_hash_value'] = 'some_hash_value'
        sot = image.Image(**example_unsupported)

        resp1 = FakeResponse(example_unsupported)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        self.assertRaises(exceptions.SDKException, sot.download, self.sess)

    def test_download_unsupported_hash_algo_falls_back_to_md5(self):
        correct_md5 = hashlib.md5(b"abc", usedforsecurity=False).hexdigest()
        example_unsupported = EXAMPLE.copy()
        example_unsupported['os_hash_algo'] = 'ancient_hash_algo'
        example_unsupported['os_hash_value'] = 'irrelevant_hash'
        example_unsupported['checksum'] = correct_md5
        sot = image.Image(**example_unsupported)

        resp1 = FakeResponse(example_unsupported)
        resp2 = FakeResponse(
            b"abc",
            headers={"Content-Type": "application/octet-stream"},
        )

        self.sess.get.side_effect = [resp1, resp2]

        rv = sot.download(self.sess)
        self.assertEqual(rv, resp2)

    def test_image_update(self):
        values = EXAMPLE.copy()
        del values['instance_uuid']
        sot = image.Image.existing(**values)
        # Let the translate pass through, that portion is tested elsewhere
        sot._translate_response = mock.Mock()

        resp = mock.Mock()
        resp.content = b"abc"
        headers = {
            'Content-Type': 'application/openstack-images-v2.1-json-patch',
            'Accept': '',
        }
        resp.headers = headers
        resp.status_code = 200
        self.sess.patch.return_value = resp

        value = [
            {"value": "fake_name", "op": "replace", "path": "/name"},
            {"value": "fake_value", "op": "add", "path": "/instance_uuid"},
        ]

        sot.name = 'fake_name'
        sot.instance_uuid = 'fake_value'
        sot.commit(self.sess)
        url = 'images/' + IDENTIFIER
        self.sess.patch.assert_called_once()
        call = self.sess.patch.call_args
        call_args, call_kwargs = call
        self.assertEqual(url, call_args[0])
        self.assertEqual(
            sorted(value, key=operator.itemgetter('value')),
            sorted(call_kwargs['json'], key=operator.itemgetter('value')),
        )

    def test_image_find(self):
        sot = image.Image()

        self.sess._get_connection = mock.Mock(return_value=self.cloud)
        self.sess.get.side_effect = [
            # First fetch by name
            FakeResponse(None, 404, headers={}, reason='dummy'),
            # Then list with no results
            FakeResponse({'images': []}),
            # And finally new list of hidden images with one searched
            FakeResponse({'images': [EXAMPLE]}),
        ]

        result = sot.find(self.sess, EXAMPLE['name'])

        self.sess.get.assert_has_calls(
            [
                mock.call(
                    'images/' + EXAMPLE['name'],
                    microversion=None,
                    params={},
                    skip_cache=False,
                ),
                mock.call(
                    '/images',
                    headers={'Accept': 'application/json'},
                    microversion=None,
                    params={'name': EXAMPLE['name']},
                ),
                mock.call(
                    '/images',
                    headers={'Accept': 'application/json'},
                    microversion=None,
                    params={'os_hidden': True},
                ),
            ]
        )

        self.assertIsInstance(result, image.Image)
        self.assertEqual(IDENTIFIER, result.id)
